functor ProofRulesDefaultFun(structure Hash: HASH
			     structure Interface: INTERFACE): PROOF_RULES_DEFAULT =
  struct
    structure Sequent = 
	SequentDefaultFun(structure Hash = Hash
			  structure Interface = Interface)

    open Sequent
    open InputLanguage
    open Typecheck

    structure Trans_SMV = TransSMVFun (structure Hash = Hash
				       structure Abstract = Abstract)

    open Hash
    open Interface
    open Options

    open SympBug
    open Str
    open Pos
    open Hash
    open ParserDefault
    open ParseTreeStruct
    open Abstract
    open Trans
    open Trans_SMV

    val trueVal = Builtin{name=True dp, Type=BoolType dp}
    val falseVal = Builtin{name=False dp, Type=BoolType dp}
    val notVal = Builtin{name=Not dp, Type=FunType(dp,BoolType dp, BoolType dp)}
    val andVal = Builtin{name=And dp, 
			 Type=FunType(dp,TupleType(dp,[BoolType dp, BoolType dp]),
				      BoolType dp)}
    val orVal = Builtin{name=Or dp, 
			Type=FunType(dp,TupleType(dp,[BoolType dp, BoolType dp]),
				     BoolType dp)}
    val equalVal = Builtin{name=Eq dp, 
			   Type=FunType(dp,TupleType(dp,[Tvar(dp,"a"),Tvar(dp,"a")]),
					BoolType dp)}
    val impliesVal = Builtin{name=Implies dp, 
			     Type=FunType(dp,TupleType(dp,[BoolType dp, BoolType dp]),
					  BoolType dp)}
    fun negFormulaTemp f = ApplTemp(BuiltinTemp notVal, f)
    fun impliesFormulaTemp(f,g) = ApplTemp(BuiltinTemp impliesVal, TupleTemp[f,g])
    fun eqExprTemp(f,g) = ApplTemp(BuiltinTemp equalVal, TupleTemp[f,g])
    fun foldOpTemp (Op,default) lst =
	let fun foldFun(f1,f2) = ApplTemp(Op,TupleTemp[f1,f2])
	    fun doFold [] = default
	      | doFold (f::lst) = List.foldr foldFun f lst
	in doFold(List.rev lst)
	end
    val foldAndTemp = foldOpTemp(BuiltinTemp andVal, BuiltinTemp trueVal)
    val foldOrTemp = foldOpTemp(BuiltinTemp orVal, BuiltinTemp falseVal)

    (* Check if the term under `id' is the constant `const' *)
    fun compareTermIDconst termHash const id =
	let val term = termID2term termHash id
	in
	    case term of
		ConstTerm c => ptEq(c, const)
	      | _ => false
	end

    (*************************************************************)
    (******************* The model check rule ********************)
    (*************************************************************)

    fun matchMC(seq: Sequent, args: RuleArg list, subOpt) =
	    
	let (* An empty dummy substitution - we don't really need one *)
	    val sigma=(case subOpt of
			   SOME sub => sub
			 | NONE => makeSubstitution(makeTermHash()))
	(* This rule should be generally applicable, at least so far,
	   so just return SOME *)
	in SOME sigma
	end

    fun translateTo "smv" options (findObject, lim, specs, vars, init, invar, next) =
	let val _ = verb options ("Translating to SMV...")
	    val (smvStr, varHash, typeHash) = 
		  SMVcode options findObject lim specs vars init invar next
	    val _ = verb options ("Translating to SMV...done")
	in (smvStr, varHash, typeHash)
	end
      | translateTo x _ _ = raise ProverError
	  ("Unsupported back-end model checker: "^x)

    val options = getOptions()
    
    val MCspecs =
	let val { limit=lim, ...} = options
	in
	    [{name="mc",
	      Type=ProverArgStringType,
	      default=DefaultValue(ProverArgString "smv")},
	     {name="limit",
	      Type=ProverArgNumberType,
	      default=DefaultValue(ProverArgNumber lim)},
	     {name="args",
	      Type=ProverArgListType ProverArgStringType,
	      default=DefaultValue(ProverArgList[])}]
	end

    fun applyMC printFun (seq: Sequent, args: RuleArg list, sigma: Substitution) =
	let fun op * (s1,s2) = Conc(s1,s2)
	    val {model=model,
		 init=init,
		 context=context,
		 invar=invar,
		 assumptions=assump,
		 conclusions=conc, ...} = seq
	    val {trans=trans,
		 findObject=findObject,
		 pvars=pvars,
		 abstraction=abs,...} = model
	    (* The back-end used to be determined in the `options', 
	       but now it's in the rule's argument *)
	    val backendMC =
		(case findArgStringValue args "mc" of
		     NONE => raise SympBug
			 ("default ProofRules/applyMC: the mc argument "
			  ^"is not a string or not defined")
		   | SOME s => s)
	    val lim =
		(case findArgNumberValue args "limit" of
		     NONE => raise SympBug
			 ("default ProofRules/applyMC: the limit argument "
			  ^"is not a number or not defined")
		   | SOME n => n)
	    val MCargs =
		(case findArgListValue args "args" of
		     NONE => raise SympBug
			 ("default ProofRules/applyMC: the `args' argument "
			  ^"is not a list or not defined")
		   | SOME lst => List.mapPartial ArgStringValue lst)
	    val verb = verb options
	    val verbStr = verbStr options
	    val lazyVerbDebug = lazyVerbDebug options
	    val verbDebug = verbDebug options
	    val lazyVerbDebugStr = lazyVerbDebugStr options
	    val trueVal = Builtin{name=True dp, Type=BoolType dp}
	    val falseVal = Builtin{name=False dp, Type=BoolType dp}
	    val andVal = Builtin{name=And dp, 
				 Type=FunType(dp,TupleType(dp,[BoolType dp, BoolType dp]),
					      BoolType dp)}
	    val orVal = Builtin{name=Or dp, 
				Type=FunType(dp,TupleType(dp,[BoolType dp, BoolType dp]),
					     BoolType dp)}
	    val impliesVal = Builtin{name=Implies dp, 
				     Type=FunType(dp,TupleType(dp,[BoolType dp, BoolType dp]),
						  BoolType dp)}
	    fun impliesFormula(f,g) = Appl(dp, impliesVal, TupleExpr(dp,[f,g]))
	    fun foldOp (Op,default) lst =
		let fun foldFun(f1,f2) = Appl(dp,Op,TupleExpr(dp,[f1,f2]))
		    fun doFold [] = default
		      | doFold (f::lst) = List.foldr foldFun f lst
		in doFold(List.rev lst)
		end
	    val foldAnd = foldOp(andVal, trueVal)
	    val foldOr = foldOp(orVal, falseVal)

	    val theorem = impliesFormula(foldAnd assump, foldOr conc)
	    val _ = verb("Specializing model to the property...")
	    val newModel = specializeModel options model [theorem]
	    val _ = verb("Specializing model to the property...done")
	    val { stateVars=vars, ... } = newModel
	    (* Check that the model is finite *)
	    val infVars = ref[]
	    val largeVars = ref[]
	    val error = ref (NONE: string option)
	    fun addError str =
		error := SOME(case !error of
				  NONE =>  str
				| SOME s => (s^"\n\n"^str))
	    fun checkVarType v =
		let val tp = getExprType findObject v
		    val size = getTypeSize options tp
		in
		    (case size of
		        InfiniteSize => infVars := (TypedExpr(dp,v,tp))::(!infVars)
		      | FiniteSize n =>
			    if n > lim then largeVars := ((TypedExpr(dp,v,tp)), n)::(!largeVars)
			    else ())
		end
	    val _ = List.app checkVarType vars
	    val _ = (case !infVars of
			 [] => ()
		       | lst => addError
			     ("The model has variables of infinite types:\n  "
			      ^(ptlist2str "\n  " lst)))
	    val _ = (case !largeVars of
			 [] => ()
		       | lst =>
			     let fun pr(pt, n) = (pt2string pt)^", size = "^(Int.toString n)
				 val strList = List.map pr lst
			     in addError("The model has variables of too large types:\n  "
					 ^(strlist2str "\n  " strList))
			     end)
	    val _ = (case !error of
			 NONE => ()
		       | SOME str => raise TransError str)
	    (* At this point, the model should be finite and small enough.
	       Continue with the translation. *)
	    val _ = verb("Translating theorems...")
	    val spec = boolExpr options findObject vars abs pvars lim theorem
	    val _ = verb("Translating theorems...done")
	    val _ = verb("Generating transition relation...")
	    val (invar, next, init) = generateModelTrans options newModel
	    val _ = verb("Generating transition relation...done")
		
	    val _ = lazyVerbDebugStr "common" 
		     (fn()=>(Str "\nINIT\n  ")*(pt2str init)*(Str "\n"))
	    val _ = lazyVerbDebugStr "common" 
		    (fn()=>(Str "\nINVAR\n  ")*(pt2str invar)*(Str "\n"))
	    val _ = lazyVerbDebugStr "common" 
		    (fn()=>(Str "\nTRANS\n  ")*(pt2str next)*(Str "\n"))
	    val _ = lazyVerbDebugStr "common" (fn()=>(Str "\nSPEC ")*(pt2str spec)*(Str "\n"))
	    (* Although called `smvStr', it is actually a program in
	       the language of whatever back-end model checker we are
	       using. *)
	    val (smvStr, varHash, typeHash) = 
		    translateTo backendMC options 
		      (findObject, lim, [spec], vars, init, invar, next)
	    val _ = verbDebugStr options "common" ((Str "\n\nSMV program:\n\n")
						   *smvStr
						   *(Str "-- end of SMV program\n\n"))
	    fun dump outs (Str s) = TextIO.output(outs,s)
	      | dump outs (Conc(x,y)) = ((dump outs x);(dump outs y))
	    fun wrapper "smv" = "smvrun"
	      | wrapper x = raise ProverError("Unsupported back-end model checker: "^x)
		(* Assume that the wrapper first reads the entire program from stdin,
		   and only then starts up the actual model checker and prints
		   something on stdout. *)
	    fun runExternalMC(smvStr,smvArgs) =
		let val proc = (Unix.execute("/bin/sh",
					     ["-c",
					      (wrapper backendMC)^" - -- "
					      ^(strlist2str " " smvArgs)])
				(* Any exception will be thrown in the model checker thread,
				   so we just exit with error *)
		               handle _ => (print("Can't execute "^backendMC);
					    raise SympExitError))
		    val (ins, outs) = (Unix.streamsOf proc
		               handle _ => raise ProverError
				   ("Can't execute/connect to "^backendMC))
		    val msg = ("Running "^backendMC^"...")
		    fun readOutput() = 
			let val str = TextIO.inputLine ins
			in
			    if str = "" then ()
			    else (printFun str; readOutput())
			end
		    fun isSuccess Posix.Process.W_EXITED = true
		      | isSuccess _ = false
		    (* Returns true if property is true *)
		    fun closeMC() = isSuccess(Unix.reap proc)
		in (verb(msg);
		    dump outs (smvStr); TextIO.closeOut outs;
		    readOutput();
		    (closeMC()) before (verb(msg^"done")))
		end
	in
	    if runExternalMC(smvStr,MCargs) then
		InferenceRuleSuccess([]: (Sequent * (SequentHints option)) list)
	    else InferenceRuleDisproven
	end handle TransError str => (printFun("\nCan't generate finite model:\n  "
					       ^str^"\n");
				      InferenceRuleFailure)

    val modelCheckRule =
	{ name="modelcheck",
	  help=("Runs an external model checker on the current sequent.\n"
		^"MC specifies the shell command to call the model checker,\n"
		^"and ARGS is the list of command line arguments to the program."),
	  params=MCspecs,
	  conclusion=NONE,
	  premisses=NONE,
	  match=SOME matchMC,
	  apply=SOME applyMC }: InferenceRule

    (*************************************************************)
    (**************** The Cone of Influence rule *****************)
    (*************************************************************)

    fun matchCOI(seq: Sequent, args: RuleArg list) =
	    
	let (* An empty dummy substitution - we don't really need one *)
	    val sigma=makeSubstitution(makeTermHash())
	(* This rule should be generally applicable, at least so far,
	   so just return SOME *)
	in SOME sigma
	end


    (* Need to extend COI to the parallel composition first *)

(*    fun applyCOI printFun (seq: Sequent, args: InferenceRuleArgument list, sigma: Substitution) =
*)
	

    (* Delete some formulas from the sequent (weakening) *)
    val deleteRule =
	{ name = "delete",
	  help = "Deletes specified formulas from the sequent",
	  params = [{name="formulas",
		     Type=ProverArgListType ProverArgFormulaNumType,
		     default=DefaultRequired}],
	  conclusion = SOME{ model=ValueTemp "model",
			     init=ValueTemp "init",
			     context=ProgramContextTemp "context",
			     invar=[ListFormulaTemp("formulas",NONE),
				    ListFormulaTemp("invar",NONE)],
			     assumptions=[ListFormulaTemp("formulas",NONE),
					  ListFormulaTemp("assump",NONE)],
			     conclusions=[ListFormulaTemp("formulas",NONE),
					  ListFormulaTemp("conc",NONE)] },
	  premisses = SOME[{ model=ValueTemp "model",
			     init=ValueTemp "init",
			     context=ProgramContextTemp "context",
			     invar=[ListFormulaTemp("invar",NONE)],
			     assumptions=[ListFormulaTemp("assump",NONE)],
			     conclusions=[ListFormulaTemp("conc",NONE)] }],
	  match = NONE,
	  apply = NONE }: InferenceRule

    (* The "A is A" axioms: when one of the formulas in the conclusion
       also appears either in the assumptions or the invariants.
       These axioms are tried on every new sequent automatically. *)

    fun initMatch(seq, args: RuleArg list, subOpt: Substitution option) =
	let val sub as (_, termHash) =
	          (case subOpt of
		       SOME sub => sub
		     | NONE => makeSubstitution(makeTermHash()))
	    val seqTerm = sequent2term(seq, SOME termHash)
	    val { model=model,
		  init=init,
		  context=context,
		  invar=invar,
		  assumptions=assump,
		  conclusions=conc, ... } = seqTerm
	    (* find a common ID in the two lists *)
	    fun findCommon ([], _) = NONE
	      | findCommon (id1::l1, l2) =
		  (case List.find(fn id2 => id1 = id2) l2 of
		       SOME id => SOME id
		     | NONE => findCommon(l1,l2))
	in
	    if List.exists (compareTermIDconst termHash trueVal) conc
		orelse List.exists (compareTermIDconst termHash falseVal) invar
		orelse List.exists (compareTermIDconst termHash falseVal) assump
	    then SOME sub
	    else
		(case findCommon(invar, conc) of
		     SOME _ => SOME sub
		   | NONE => (case findCommon(assump, conc) of
				  SOME _ => SOME sub
				| NONE => NONE))
	end

    fun initApply printFun (seq, args, sub) =
	(case initMatch(seq, args, SOME sub) of
	     SOME _ => InferenceRuleSuccess[]
	   | NONE => InferenceRuleFailure)

    val initRule =
	{ name = "init",
	  help = ("Rule for the axioms `A => A', `FALSE => A', and `A => TRUE'.\n"
		  ^"Completes the proof if the invariants or assumptions have common\n"
		  ^"formulas with the conclusion part, or if the sequent is\n"
		  ^"trivially true.\n"),
	  params = [],
	  conclusion = NONE,
	  premisses = NONE,
	  match = SOME initMatch,
	  apply = SOME initApply }: InferenceRule

    (* Make a copy of a formula in the sequent *)
    val copyRule =
	{ name = "copy",
	  help = "Makes a copy of the specified formula",
	  params = [{name = "formula",
		     Type = ProverArgFormulaNumType,
		     default = DefaultRequired}],
	  conclusion = SOME{ model = ValueTemp "model",
			     init = ValueTemp "init",
			     invar = [ListFormulaTemp("formula",NONE),
				      ListFormulaTemp("invar",NONE)],
			     context=ProgramContextTemp "context",
			     assumptions = [ListFormulaTemp("formula",NONE),
					    ListFormulaTemp("assump",NONE)],
			     conclusions = [ListFormulaTemp("formula",NONE),
					    ListFormulaTemp("conc",NONE)] },
	  premisses = SOME[{ model = ValueTemp "model",
			     init = ValueTemp "init",
			     context=ProgramContextTemp "context",
			     invar = [ListFormulaTemp("formula",NONE),
				      ListFormulaTemp("formula",NONE),
				      ListFormulaTemp("invar",NONE)],
			     assumptions = [ListFormulaTemp("formula",NONE),
					    ListFormulaTemp("formula",NONE),
					    ListFormulaTemp("assump",NONE)],
			     conclusions = [ListFormulaTemp("formula",NONE),
					    ListFormulaTemp("formula",NONE),
					    ListFormulaTemp("conc",NONE)] }],
	  match = NONE,
	  apply = NONE }: InferenceRule

    (* Copy formulas from the invariant set to the assumptions *)
    val useInvarRule =
	{ name = "useinvar",
	  help = "Copies a specified formula from the invariant to the assumption part",
	  params = [{name = "formulas",
		     Type = ProverArgListType ProverArgFormulaNumType,
		     default = DefaultRequired}],
	  conclusion = SOME{ model = ValueTemp "model",
			     init = ValueTemp "init",
			     context=ProgramContextTemp "context",
			     (* The `formulas' var should match all the formulas by default *)
			     invar = [ListFormulaTemp("formulas",NONE), ListFormulaTemp("invar",NONE)],
			     assumptions = [ListFormulaTemp("assump",NONE)],
			     conclusions = [ListFormulaTemp("conc",NONE)] },
	  premisses = SOME[{ model = ValueTemp "model",
			     init = ValueTemp "init",
			     context=ProgramContextTemp "context",
			     invar = [ListFormulaTemp("formulas",NONE),
				      ListFormulaTemp("invar",NONE)],
			     assumptions = [ListFormulaTemp("assump",NONE),
					    ListFormulaTemp("formulas",NONE)],
			     conclusions = [ListFormulaTemp("conc",NONE)] }],
	  match = NONE,
	  apply = NONE }: InferenceRule

    (* The `normal' cut rule, adds a formula to the assumptions and then to the conclusions.
       The cut for the invariant is more complex and requires modification of the
       formula. *)
    val caseRule =
	{ name = "case",
	  help = "Split cases on whether a given formula is true or false",
	  params = [{name = "formula",
		     (* The user supplies his own new formula *)
		     Type = ProverArgObjectType,
		     default = DefaultRequired}],
	  conclusion = SOME{ model = ValueTemp "model",
			     init = ValueTemp "init",
			     context=ProgramContextTemp "context",
			     invar = [ListFormulaTemp("invar",NONE)],
			     assumptions = [ListFormulaTemp("assump",NONE)],
			     conclusions = [ListFormulaTemp("conc",NONE)] },
	  premisses = SOME[{ model = ValueTemp "model",
			     init = ValueTemp "init",
			     context=ProgramContextTemp "context",
			     invar = [ListFormulaTemp("invar",NONE)],
			     (* Add the new formula to the assumptions *)
			     assumptions = [FormulaTemp "formula",
					    ListFormulaTemp("assump",NONE)],
			     conclusions = [ListFormulaTemp("conc",NONE)] },
			   { model = ValueTemp "model",
			     init = ValueTemp "init",
			     context=ProgramContextTemp "context",
			     invar = [ListFormulaTemp("invar",NONE)],
			     assumptions = [ListFormulaTemp("assump",NONE)],
			     (* Add the new formula to the conclusions *)
			     conclusions = [FormulaTemp "formula",
					    ListFormulaTemp("conc",NONE)] }],
	  match = NONE,
	  apply = NONE }: InferenceRule

    (* $\forall Case_R$ rule *)
    val forallCaseRule =
	{ name = "forallcase",
	  help = "Splitting cases on all possible values of the given term",
	  params = [{name="var",
		     Type=ProverArgObjectType,
		     default=DefaultRequired},
		    {name="term",
		     Type=ProverArgObjectType,
		     default=DefaultRequired},
		    {name="formula",
		     Type=ProverArgFormulaNumType,
		     default=DefaultMatch NONE}],
	  conclusion = SOME{ model = ValueTemp "model",
			     init = ValueTemp "init",
			     context=ProgramContextTemp "context",
			     invar = [ListFormulaTemp("invar",NONE)],
			     (* Add the new formula to the assumptions *)
			     assumptions = [ListFormulaTemp("assump",NONE)],
			     conclusions = [FormulaTemp "formula",
					    ListFormulaTemp("conc",NONE)] },
	  premisses = SOME[{ model = ValueTemp "model",
			     init = ValueTemp "init",
			     context=ProgramContextTemp "context",
			     invar = [ListFormulaTemp("invar",NONE)],
			     (* Add the new formula to the assumptions *)
			     assumptions = [ListFormulaTemp("assump",NONE)],
			     conclusions = [ListFormulaTemp("conc",NONE),
					    ForallTemp([FormulaTemp "var"],
						       impliesFormulaTemp
						         (eqExprTemp(FormulaTemp "var",
								     FormulaTemp "term"),
							  FormulaTemp "formula"),
							FormulaTemp "context")] }],
	  match = NONE,
	  apply = NONE }: InferenceRule

    (* Flatten rule: eliminate all conjunctions in the invariants and
       assumptions and disjunctions in the conclusions.  For now, it
       doesn't take any arguments. *)

    fun flattenMatch(seq, args: RuleArg list, subOpt) =
	let val sub as (_, termHash) =
	          (case subOpt of
		       SOME sub => sub
		     | NONE => makeSubstitution(makeTermHash()))
	    val seqTerm = sequent2term(seq, SOME termHash)
	    val { model=model,
		  init=init,
		  context=context,
		  invar=invar,
		  assumptions=assump,
		  conclusions=conc, ... } = seqTerm
	    (* Flag whether the sequent has changed *)
	    val changed = ref false
	    (* Extract arguments of "and", "or", "not", and "implies" *)
	    fun splitArgs id =
		(case termID2term termHash id of
		     TupleTerm lst => SOME lst
		   | _ => SOME[id])

	    fun splitOp Op id =
		let val term = termID2term termHash id
		in
		    case term of
			ApplTerm(opId, argId) =>
			    (if compareTermIDconst termHash Op opId then splitArgs argId
			     else NONE)
		      | _ => NONE
		end

	    (* In the invariant, only split conjunctions.  Negations
	       are kept in to preserve completeness. *)
	    fun splitInvar id = Option.map(fn l => (l, [], [])) (splitOp andVal id)

	    (* Assumptions are split on conjunctions and negations *)
	    fun splitAssump id =
		  (case splitOp andVal id of
		       (* Keep conjuncts in the assumptions *)
		       SOME l => SOME([], l, [])
		     | NONE => 
			   (case splitOp notVal id of
				(* Move un-negated formulas to conclusions *)
				SOME l => SOME([], [], l)
			      | NONE => NONE))

	    fun splitConc id =
		  (case splitOp orVal id of
		       SOME l => SOME([], [], l)
		     | NONE => 
			   (case splitOp impliesVal id of
				SOME[id1,id2] => SOME([], [id1], [id2])
			      | NONE =>
				    (case splitOp notVal id of
					 SOME l => SOME([], l, [])
				       | NONE => NONE)
			      | SOME l => raise SympBug
				  ("ProofRulesDefault/flattenMatch/splitConc: "
				   ^"splitImplies returned bad list:\n  ["
				   ^(strlist2str ", " (List.map termID2str l))
				   ^"]")))

	    (* 6 arguments: 3 lists of formulas to split, 3 lists of
	       processed formulas.  War plan: look at the first
	       formula in each "to do" list, and if it splits, put the
	       parts in the corresponding "to do" lists again; if it
	       doesn't split, move it to the corresponding processed
	       list *)

	    fun loop([], [], [], invarDone, assumpDone, concDone) =
		  (List.rev invarDone, List.rev assumpDone, List.rev concDone)
	      | loop(x::invar, assump, conc, invarDone, assumpDone, concDone) =
		  (case splitInvar x of
		       NONE => loop(invar, assump, conc, x::invarDone, assumpDone, concDone)
		     | SOME(l1,l2,l3) => (loop(l1@invar, l2@assump, l3@conc,
					       invarDone, assumpDone, concDone)
					  before changed := true))
	      | loop (invar, x::assump, conc, invarDone, assumpDone, concDone) =
		  (case splitAssump x of
		       NONE => loop(invar, assump, conc, invarDone, x::assumpDone, concDone)
		     | SOME(l1,l2,l3) => (loop(l1@invar, l2@assump, l3@conc,
					       invarDone, assumpDone, concDone)
					  before changed := true))
	      | loop (invar, assump, x::conc, invarDone, assumpDone, concDone) =
		  (case splitConc x of
		       NONE => loop(invar, assump, conc, invarDone, assumpDone, x::concDone)
		     | SOME(l1,l2,l3) => (loop(l1@invar, l2@assump, l3@conc,
					       invarDone, assumpDone, concDone)
					  before changed := true))
	    val (invar, assump, conc) = loop(invar, assump, conc, [], [], [])
	    val invarValue = ListValue(List.map (fn id => ObjectValue id) invar)
	    val assumpValue = ListValue(List.map (fn id => ObjectValue id) assump)
	    val concValue = ListValue(List.map (fn id => ObjectValue id) conc)
	    fun updateSub() = 
		let val sOpt = updateSubstitution("model", ModelValue model) sub
		    val sOpt = Option.mapPartial
			         (updateSubstitution("init", InitValue init)) sOpt
		    val sOpt = Option.mapPartial
			         (updateSubstitution("context", ProgramContextValue context)) sOpt
		    val sOpt = Option.mapPartial
			         (updateSubstitution("invar", invarValue)) sOpt
		    val sOpt = Option.mapPartial
			         (updateSubstitution("assump", assumpValue)) sOpt
		    val sOpt = Option.mapPartial
			         (updateSubstitution("conc", concValue)) sOpt
		in
		    sOpt
		end
	in
	    if !changed then updateSub() else NONE
	end

    fun flattenApply printFun (seq, args, sigma) =
	let val seqTemp = { model = ValueTemp "model",
			    init = ValueTemp "init",
			    context=ProgramContextTemp "context",
			    invar = [ListFormulaTemp("invar",NONE)],
			    assumptions = [ListFormulaTemp("assump",NONE)],
			    conclusions = [ListFormulaTemp("conc",NONE)] }
	    val newSub = flattenMatch(seq, args, SOME sigma)
	    val newSeq = Option.mapPartial(fn sub=>substSequent sub seqTemp) newSub
	in
	    case newSeq of
		SOME seq => InferenceRuleSuccess[(seq,NONE)]
	      | NONE => InferenceRuleFailure
	end

    val flattenRule =
	{ name = "flatten",
	  help = ("Propositionally flattens the sequent.  That is, applies disjunction on"
		  ^"the right, conjunction on the left, and rules for negation as much as"
		  ^"possible."),
	  params = [],
	  conclusion = NONE,
	  premisses = NONE,
	  match = SOME flattenMatch,
	  apply = SOME flattenApply }: InferenceRule

			
    fun proofRules() = [modelCheckRule,
			deleteRule,
			initRule,
			(* initInvarRule,
			initAssumpRule, *)
			copyRule,
			useInvarRule,
			caseRule,
			forallCaseRule,
			flattenRule]

  end
