(* The default implementation of the PROOF_SYSTEM abstract interface *)

functor ProofSystemDefaultFun(structure Hash: HASH
			      structure Interface: INTERFACE): PROOF_SYSTEM =
  struct
    structure ProofRules = 
	ProofRulesDefaultFun(structure Hash = Hash
			     structure Interface = Interface)

    open Interface
    open ProofRules
    open Sequent
    open SequentCommon
    open Hash
    open SympBug

    local
	val rules = makeHashDefault(op =, fn x=>x): (string, InferenceRule) Hash
    in
	fun findRuleByName str = findHash(rules, str)
	fun addNewRule r = (insertHashDestructive(rules, getRuleName r, r); ())

	(* Initialize the set of inference rules from the ProofRules module *)
	fun initProofSystem () = List.app addNewRule (proofRules())
	fun getAllRules() = List.map #2 (hash2any(fn x => x)(fn x => x) rules)

	fun getInitRule() = 
	    (case findRuleByName "init" of
		 SOME r => r
	       | NONE => raise SympBug
		     ("ProofSystemDefault/getInitRule: no \"init\" rule"))

	(* Apply the rule, and check if the resulting sequents are trivially
	   true by applying the init rule.  All sequents that are proven this way
	   are discarded from the result. *)

	fun applyRule printFun (seq, context) (rule, args, substOpt) =
	    let val debug = verbDebug(getOptions()) "applyRule"
		val res = Sequent.applyRule printFun (seq, context) (rule, args, substOpt)
		val _ = debug("\napplyRule: checking for trivially true sequents\n")
		(* Annihilate the sequent if it's trivially true *)
		fun checkSeq (pair as (seq, _)) =
		    (case Sequent.applyRule printFun (seq, context) (getInitRule(), [], NONE) of
			 InferenceRuleSuccess [] => NONE
		       | _ => SOME pair)
		(* Clean up all trivial sequents *)
		fun checkSeqs(InferenceRuleSuccess seqs) =
		      InferenceRuleSuccess(List.mapPartial checkSeq seqs)
		  | checkSeqs x = x
	    in
		checkSeqs res
	    end

    end

  end
