functor TransSMVFun (structure Hash : HASH
		     structure Abstract : ABSTRACT): TRANS_SMV =

  struct

    structure Hash = Hash
    structure Abstract = Abstract

    open Str
    open Hash
    open Abstract
    open Pos
    open ParserDefault
    open ParseTreeStruct
    open Interface
    open Options

    type VarSMVMapping = (ParseTree, Str) Hash
    type DefineMapping = (ParseTree, Str) Hash
    type TypeValueMapping = (ParseTree, Str list) Hash

    fun ptListEq (x, y) = List.all (fn n => (ptEq (List.nth (x, n),
						   List.nth (y, n)))
				    handle Subscript => false)
      (List.tabulate (List.length x, fn i => i))

    (* Redefine name generator _not_ to include `!' symbols *)
    fun newName () = "X_"^(Int.toString(newNumber()))

    fun makeMapping () = makeHashDefault (ptEq, pt2string) 
    fun makeTVMapping () = makeHashDefault (ptEq, pt2string) 
    fun addMapDestructive(map, v, smv) = 
	(lazyVerbDebug (getOptions()) "addMap" (fn()=>"addMap("^(pt2string v)^")\n");
	 insertHashDestructive(map, v, smv))

    fun getMap options (map, v) =
	let val debug = lazyVerbDebug options "getMap"
	    val _ = debug(fn()=>"\ngetMap("^(pt2string v)^", ["
			  ^(strlist2str ",\n          "
			    (List.map(fn(v,v')=>(pt2string v)^" -> "^(Str2string v'))
			     (hash2any(fn x=>x)(fn x=>x) map)))
			  ^"])\n")
	in (case findHash (map, v) of
		SOME s => s
	      | NONE => raise SympBug
		 ("trans_smv:  variable not in hash: "^(pt2stringDebug v)))
	end

    fun getTVMap options (map, v) =
	let val debug = lazyVerbDebug options "getMap"
	    val _ = debug(fn()=>"\ngetTVMap("^(pt2string v)^", ["
			  ^(strlist2str ",\n          "
			    (List.map(fn(v,lst)=>(pt2string v)^" -> "
				      ^(Str2string (Strlist2Str ", "  lst)))
			     (hash2any(fn x=>x)(fn x=>x) map)))
			  ^"])\n")
	in (case findHash (map, v) of
		SOME s => s
	      | NONE => raise SympBug
		 ("trans_smv:  variable not in hash: "^(pt2stringDebug v)))
	end

    fun isComplex (Appl (_, Builtin{name=And _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=Or _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=Implies _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=Eq _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=NotEq _, ...}, _)) = true
      | isComplex (Appl (_, Builtin{name=Not _, ...}, _)) = true
      | isComplex x = false


    val bottom = Conc (Str (newName ()), Str "bot")
    val truth = Conc (Str (newName ()), Str "true")
    val falsity = Conc (Str (newName ()), Str "false")
    val eqTrue = Conc (Str " = ", truth)
    val eqFalse = Conc (Str " = ", falsity)
    val eqBottom = Conc (Str " = ", bottom)
    val thenTrue = Conc (Str " : ", Conc (truth, Str ";"))
    val thenFalse = Conc (Str " : ", Conc (falsity, Str ";"))
    val thenBottom = Conc (Str " : ", Conc (bottom, Str ";"))
    val otherwise = (Str "\n  1")
    val || = (Str " | ")
    val & = (Str " & ")
    val dummyBool = Conc (Str (newName ()), Str "ensureBoolean")

    (* Functions to translate 1- and 2-arity functions.
       There should be NO trailing `;' at the end. *)

    local fun op * (x,y) = Conc(x,y)
    in 
      fun codeFor1 (Not _, s) = (Str "!(") * s * (Str ")")
	| codeFor1 (Ag _, s) = (Str "AG(") * s * (Str ")")
	| codeFor1 (Af _, s) = (Str "AF(") * s * (Str ")")
	| codeFor1 (Eg _, s) = (Str "EG(") * s * (Str ")")
	| codeFor1 (Ef _, s) = (Str "EF(") * s * (Str ")")
	| codeFor1 (Ax _, s) = (Str "AX(") * s * (Str ")")
	| codeFor1 (Ex _, s) = (Str "EX(") * s * (Str ")")
	| codeFor1 (Object {def = d, ...}, s) = codeFor1 (d, s)
	| codeFor1 (ObjectInst {obj = d, ...}, s) = codeFor1 (d, s)
	| codeFor1 (Builtin {name=n, ...}, s) = codeFor1 (n, s)
	| codeFor1 (x, s) = raise SympBug 
	  ("trans_SMV:  Unexpected arity-1 operator in formula:\n  "
	   ^(pt2string x)^"("^(Str2string s)^")")
	  
      fun codeFor2 (And _, s1, s2) = (Str "(") * s1 * & * s2 * (Str ")")
	| codeFor2 (Or _, s1, s2) = (Str "(") * s1 * || * s2 * (Str ")")
	| codeFor2 (Implies _, s1, s2) = (Str "(") * s1 * (Str " -> ") * s2 * (Str ")")
	| codeFor2 (Eq _, s1, s2) = (Str "(") * s1 * (Str " = ") * s2 * (Str ")")
	| codeFor2 (NotEq _, s1, s2) = (Str "! (") * s1 * (Str " = ") * s2 * (Str ")")
	| codeFor2 (Au _, s1, s2) = (Str "(") * s1 * (Str "AU") * s2 * (Str ")")
	| codeFor2 (Eu _, s1, s2) = (Str "(") * s1 * (Str "EU") * s2 * (Str ")")
	| codeFor2 (Ar _, s1, s2) = (Str "(") * s1 * (Str "AR") * s2 * (Str ")")
	| codeFor2 (Er _, s1, s2) = (Str "(") * s1 * (Str "ER") * s2 * (Str ")")
	| codeFor2 (Object {def = d, ...}, s1, s2) = codeFor2 (d, s1, s2)
	| codeFor2 (ObjectInst {obj = d, ...}, s1, s2) = codeFor2 (d, s1, s2)
	| codeFor2 (Builtin {name=n, ...}, s1, s2) = codeFor2 (n, s1, s2)
	| codeFor2 (x, s1, s2) = raise SympBug 
	  ("trans_SMV:  Unexpected arity-2 operator in formula:\n  "
	   ^(pt2string x)^"("^(Str2string s1)^", "^(Str2string s1)^")")

      (* Same functions for 3-valued logic *)
      fun codeFor1val3 (Not _, s) = 
	  let
	      val c1 = (Str "  ") * s * eqTrue * thenFalse
	      val c2 = (Str "  ") * s * eqFalse * thenTrue
	      val c3 = otherwise * thenBottom
	  in
	      (Str "case\n") * c1 * c2 * c3 * (Str "\nesac")
	  end
	| codeFor1val3 (x, s) = raise SympBug 
	  ("trans_SMV:  Unexpected arity-1 operator in formula:\n  "
	   ^(pt2string x)^"("^(Str2string s)^")")
	  
      fun codeFor2val3 (And _, s1, s2) = 
	  let
	    val c1 = (Str "  ") * s1 * eqFalse * || * s2 * eqFalse * thenFalse
	    val c2 = (Str "\n  ") * s1 * eqTrue * & * s2 * eqTrue * thenTrue
	    val c3 = otherwise  * thenBottom
	  in
	    (Str "case\n") * c1 * c2 * c3 * (Str "\nesac")
	  end
	| codeFor2val3 (Or _, s1, s2) = 
	  let
	    val c1 = (Str "  ") * s1 * eqTrue * || * s2 * eqTrue * thenTrue
	    val c2 = (Str "\n  ") * s1 * eqFalse * & * s2 * eqFalse * thenFalse
	    val c3 = otherwise  * thenBottom
	  in
	    (Str "case\n") * c1 * c2 * c3 * (Str "\nesac")
	  end	
	| codeFor2val3 (Implies _, s1, s2) = 
	  let
	    val c1 = (Str "  ") * s1 * eqFalse * || * s2 * eqTrue * thenTrue
	    val c2 = (Str "\n  ") * s1 * eqTrue * & * s2 * eqFalse * thenFalse
	    val c3 = otherwise * thenBottom
	  in
	    (Str "case\n") * c1 * c2 * c3 * (Str "\nesac")
	  end
	| codeFor2val3 (Eq _, s1, s2) = s1 * (Str " = ") * s2
	| codeFor2val3 (NotEq _, s1, s2) = (Str "! (") * s1 * (Str " = ") * s2 * (Str ")")
	| codeFor2val3 (x, s1, s2) = raise SympBug 
	  ("trans_SMV:  Unexpected arity-2 operator in formula:\n  "
	   ^(pt2string x)^"("^(Str2string s1)^", "^(Str2string s1)^")")

    end
(*    fun buildMap (formula, varMap) =
      let
	val map = makeMapping ()
	val str = ref (Str "")
	fun build (a as (Appl (_, Builtin{name=n, ...}, 
			       TupleExpr (_, [v1, v2])))) =
	  let
	    val v1c = isComplex v1
	    val v2c = isComplex v2
	  in
	    if v1c then
	      (build v1;
	       if v2c then
		 build v2
	       else
		 ())
	    else
	      if v2c then
		build v2
	      else
		let
		  val defName = Str (newName ())
		in
		  (addMapDestructive (map, a, defName);
		   str := Conc (Str "\nDEFINE ",
				Conc (defName, 
				      Conc (Str " := ", 
					    Conc
					    (codeFor2 (n, 
						       getMap (varMap, v1),
						       getMap (varMap, v2)),
					     !str)))))
		end
	  end
	  | build (a as (Appl (_, Builtin{name=n, ...}, v))) =
	  if isComplex v then
	    build v
	  else
	    let
	      val defName = Str (newName ())
	    in
	      (addMapDestructive (map, a, defName);
	       str := Conc (Str "\nDEFINE ",
			    Conc (defName,
				  Conc (Str " := ",
					Conc
					(codeFor1 (n, getMap (varMap, v)),
					 !str)))))
	    end
	  | build v = ()
      in
	(!str, map)
      end
*)

    fun formulaStr options (varMap, formula) =
      let val debug = lazyVerbDebug options "formulaStr"
	  fun op * (x,y) = Conc(x,y)
	fun formStr a =
	    (case findHash(varMap, a) of
		 SOME x => x
	       | NONE => 
		  (case a of
		       (Appl (_, x, TupleExpr (_, [v1, v2]))) =>
			   codeFor2 (x, formStr v1, formStr v2)
		     | (Appl (_, x, v)) => codeFor1 (x, formStr v)
		     | (Next v) => (Str "next(")*(formStr v)*(Str ")")
		     | (Builtin{name=True _, ...}) => truth
		     | (Builtin{name=False _, ...}) => falsity
		     | (Builtin{name=Undefined _, ...}) => bottom
		     | True2 => Str "1"
		     | False2 => Str "0"
		     | x => raise SympBug
			 ("formulaStr: unexpected expression: "^(pt2stringDebug x))))
	val _ = debug(fn()=>"\nformulaStr("^(pt2string formula)^")[\n")
	val _ = debug(fn()=>"varMap = ["
		      ^(strlist2str ",\n          "
			(List.map(fn(v,v')=>(pt2string v)^" -> "^(Str2string v'))
			 (hash2any(fn x=>x)(fn x=>x) varMap)))
		      ^"]\n")
	val str = formStr formula
	val _ = debug(fn()=>"\nformulaStr => "^(Str2string str)^"]\n")
      in
	  (Str "", str)
      end

    (* 3-valued variant of the translation *)
    fun formulaStr3val options (varMap, formula) =
      let val debug = lazyVerbDebug options "formulaStr"
	  fun op * (x,y) = Conc(x,y)
	fun formStr (a as (Appl (_, Builtin{name=n, ...},
				 TupleExpr (_, [v1, v2]))), d) =
	       let
		 val (nd1, s1) = formStr (v1, d)
		 val (nd2, s2) = formStr (v2, d)
		 val s1' = Str (newName ())
		 val s2' = Str (newName ())
		 val defs = (Str "\nDEFINE ")*s1'*(Str " := ")*s1
		           *(Str ";\nDEFINE ")*s2'*(Str " := ")*s2*(Str ";\n")
	       in
		 (defs * nd1 * nd2 * d,
		  codeFor2 (n, s1', s2'))
	       end
	  | formStr (a as (Appl (_, Builtin{name=n, ...}, v)), d) =
	    let
	      val (nd, s) = formStr (v, d)
	      val s' = Str (newName ())
	      val defs = (Str "\nDEFINE ") * s' * (Str " := ") * s * (Str ";\n")
	    in
	      (defs *nd * d, codeFor1 (n, s'))
	    end
	  | formStr (a as (Next v), d) = 
	      let val (nd, s) = formStr(v,d)
	      in (nd, (Str "next(")*s*(Str ")"))
	      end
	  | formStr (a as (Builtin{name=True _, ...}), d) = (d, truth)
	  | formStr (a as (Builtin{name=False _, ...}), d) = (d, falsity)
	  | formStr (a as (Builtin{name=Undefined _, ...}), d) = (d, bottom)
	  | formStr (v, d) = (d, getMap options (varMap, v))
	val _ = debug(fn()=>"\nformulaStr("^(pt2string formula)^")[\n")
	val _ = debug(fn()=>"varMap = ["
		      ^(strlist2str ",\n          "
			(List.map(fn(v,v')=>(pt2string v)^" -> "^(Str2string v'))
			 (hash2any(fn x=>x)(fn x=>x) varMap)))
		      ^"]\n")
	val (defs, str) = formStr (formula, Str "")
	val _ = debug(fn()=>"\nformulaStr => "^(Str2string str)^"]\n")
      in
	  (defs, str)
      end
		    
    fun SMVcode options findObject limit specList varList initBool invarBool transBool =
      let
	val debug = lazyVerbDebug options "SMVcode"
	val booleanSeen = ref false
	fun op * (x,y) = Conc(x,y)
	val formulaStr = formulaStr options
	val map = makeMapping ()
	val _ = (addMapDestructive (map, True(dp), truth);
		 addMapDestructive (map, False(dp), falsity))
	val TVmap = makeTVMapping ()
	fun innerTypeStr v = 
	  let 
	    val tvals = getTVMap options (TVmap, v)
	  in
	    Strlist2Str ", " tvals
	    (* List.foldr (fn (x, y) =>
			Conc (y, Conc (Str ", ", x))) (hd tvals) (tl tvals) *)
	  end
	fun typeStr v =  (Str " : {") *  (innerTypeStr v) * (Str "}")
	val _ = debug(fn()=>"\n-- SMVcode: generating types for vars:\n  ["
		      ^(ptlist2str "," varList)^"]\n")
	val str = Str "\nMODULE main\n\n"
	fun doVar v =
	    let val _ = addMapDestructive (map, v, (Str(newName ())))
		val t = getExprType findObject v
		val tvals = getTypeValues options limit t
		val _ = (case t of 
			     BoolType (_) => booleanSeen := true
			   | _ => ())
	    in
		case t of
		    AbstractType (alist) =>
			let val names = List.map 
				 (fn name => 
				  case (findHash (map, name)) of
				      SOME n => n
				    | NONE => Str (newName ())) alist
			in
			     (addMapDestructive (TVmap, v, bottom :: names);
			      List.app (fn n =>
					case (findHash (map, List.nth (alist, n))) of
					    NONE => (addMapDestructive 
						     (map, List.nth (alist, n),
						      List.nth (names, n));
						     ())
					  | SOME _ => ())
			      (List.tabulate (List.length alist, fn x => x)))
			end
		  | _ => 
			 (case tvals of
			      SOME vlist =>
				  let val names = List.map 
				      (fn name => 
				           case (findHash (map, name)) of
					       SOME n => n
					     | NONE => Str (newName ())) vlist
				  in
				      (addMapDestructive (TVmap, v, bottom::names);
				       List.app (fn n =>
						 case (findHash (map, List.nth (vlist, n))) of
						     NONE =>
							 (addMapDestructive 
							  (map, 
							   List.nth (vlist, n),
							   List.nth (names, n));
							  ())
						   | SOME _ => ())
				       (List.tabulate (List.length vlist, fn x => x)))
				  end
			    | NONE => raise ProverError
				  ("Infinite type in translation to SMV:\n  Type: "
				   ^(pt2stringDebug t)
				   ^"\n  Expr: "^(pt2string v)))
	    end
	val _ = List.app doVar varList
	val _ = debug(fn()=>"\n-- SMVcode: generating var mapping...\n")
	val comments = Strlist2Str "\n"
	  ((Str "-- SMV Code Automatically Generated by SyMP\n\n--Variable mappings:\n")
	   ::(List.map 
	      (fn v => (Str "-- ")*(pt2str v)*(Str " -> ")
	              *(getMap options (map, v))*(typeStr v))
	      varList))
	val comments = comments
	    *(Str "\n\n--Mapping of all names (including variables)\n")
	    *(Strlist2Str "\n"
	      (List.map(fn(v,str)=>(Str "-- ") * v * (Str " -> ")* str)
	       ((Str "undefined", bottom)::(hash2any pt2str (fn x=>x) map))))

	val _ = debug(fn()=>"\n-- SMVcode: vars and their mappings:\n  ["
		      ^(strlist2str ",\n   "
			(List.map(fn v=>(pt2stringDebug v)^" -> "
				  ^(Str2string (getMap options (map, v)))
				  ^" : {"
				  ^(Str2string (Strlist2Str ", " (getTVMap options (TVmap, v))))
				  ^"}") varList))
		      ^"]\n")
	val _ = debug(fn()=>"\n-- SMVcode: generating VAR declarations...\n")
	val str = comments * str
	val str = str * (Strlist2Str "\n  "
			 (((Str ("\n\nVAR\n  ")) *
			   (if (!booleanSeen) then (Str "")
			    else
			      (dummyBool * (Str ": {") * bottom * (Str ", ") * truth 
			       * (Str", ") * falsity * (Str "};\n  "))))
			  ::(List.map (fn v => (getMap options (map, v))
				              *(typeStr v)*(Str ";"))
			     varList)))
	val _ = debug(fn()=>"\n-- SMVcode: generating formula for INIT...\n")
	val (initDef, initForm) = formulaStr (map, initBool)
	val initStr = initDef * (Str "\nINIT\n  ") * initForm * (Str "\n")
	val _ = debug(fn()=>"\n-- SMVcode: generating formula for INVAR...\n")
	val (invarDef, invarForm) = formulaStr (map, invarBool)
	val invarStr = invarDef * (Str "\nINVAR\n  ") * invarForm * (Str "\n")
	val _ = debug(fn()=>"\n-- SMVcode: generating formula for TRANS...\n")
	val (transDef, transForm) = formulaStr (map, transBool)
	val transStr = transDef * (Str "\nTRANS\n  ") * transForm * (Str "\n")
	val specStr = List.foldr Conc (Str "") 
	  (List.map
	   (fn s => 
	    let
	      val (specDef, specForm) = formulaStr (map, s)
	    in
	      (Str "\nSPEC\n  ") * specForm * (Str "\n")
	    end) specList)
(*	val _ = debug(fn()=>"\nSMVcode: generating formula for INVAR...\n")
	val (invarDef, invarMap) = buildMap (invarBool, map)
	val invarForm = formulaStr (map, invarBool)
	val invarStr = Conc 
	  (initDef,
	   Conc (Str "DEFINE N := ",
		 Conc (invarForm,
		       Conc (Str "\nINVAR (N = ",
			     Conc (bottom, Str "& (error = 1)) | N = 1")))))
	val _ = debug(fn()=>"\nSMVcode: generating formula for TRANS...\n")
	val (transDef, transMap) = buildMap (transBool, map)
	val transForm = formulaStr (transMap, map, transBool)
	val transStr = Conc 
	  (initDef,
	   Conc (Str "DEFINE T := ",
		 Conc (transForm,
		       Conc (Str "\nTRANS (T = ",
			     Conc (bottom, Str "& (error = 1)) | T = 1")))))
	  *)
	(* val _ = debug(fn()=>"\n-- SMVcode: generating formula for next(error)...\n")
	val errorStr = (Str ("ASSIGN\n"
			     ^"  init(error) := 0;\n"
			     ^"  next(error) :=\n"
			     ^"    case\n"
			     ^"      error_invar | error_trans: 1;\n"
			     ^"      1: error;\n"
			     ^"    esac;\n")) *)
	val _ = debug(fn()=>"\n-- SMVcode: Translation finished.\n")
	val str =  str *  initStr * invarStr * transStr * specStr (* * errorStr *)
      in
	(str, map, TVmap)
      end

  end
