functor TransPrepFun(structure TransCommon: TRANS_COMMON): TRANS_PREP =
  struct
    structure TransCommon = TransCommon

    open TransCommon

    open Hash
    open Pos
    open Abstract
    open Evaluate
    open ParserDefault
    open ParseTreeStruct
    open Interface
    open Options

(*      val dummyAVT = NormalAsstTree({norm=[],next=[],init=[]}, []) *)

    (* l1 - l2 as sets *)
    fun setMinus eq (l1, l2) = List.filter(fn x=>not(List.exists(fn y=>eq(x,y)) l2)) l1

    val unionPt = union ptEq

    (* For a list of assignments generate an AsstVarsTree.  On the
       way, check for circular and multiple assignments, and if found
       any, raise TransError.  Also, check that all the `stateVars'
       are assigned. *)

    fun collectAsstVars options findObject lim (stateVars, assts) = 
	let val cone = makeCone() (* Global cone, update destructively *)
	    val pvars = makePrimitiveVars()
	    val debug = lazyVerbDebug options
	    (* Make our duly pray to the ObjectInst, just out of superstition *)
	    (* val stateVars = List.map stripObjectInst stateVars *)
	    val _ = debug "collectAsstVars"
		(fn()=>"\ncollectAsstVars(["^(ptlist2strDebug "," stateVars)
		 ^"],\n                ["
		 ^(ptlist2str ",\n                 " assts)
		 ^"])[\n")
	    val addPrim = addPrim pvars
	    val addToCone = addToCone options findObject pvars lim cone
	    val addToCones = addToCones options findObject pvars lim cone
	    (* A string with an error message, or NONE. *)
	    val error = ref(NONE)
	    (* Lists of problematic variables: *)
	    (* Multiply assigned vars *)
	    val multiplyAssignedNormal = ref(NONE)
	    val multiplyAssignedNext = ref(NONE)
	    val multiplyAssignedInit = ref(NONE)
	    (* Variables appearing in too many places *)
	    val normalAndNext = ref(NONE)
	    val normalAndInit = ref(NONE)
	    (* Variables appearing in too few places *)
	    val initButNotNext = ref(NONE)
	    val nextButNotInit = ref(NONE)
	    val unassigned = ref(NONE)
	    fun addError str =
		error := SOME(case !error of
				  NONE =>  str
				| SOME s => (s^"\n\n"^str))
	    fun addErrorVars varList vars =
		varList := SOME(case !varList of
				    NONE => vars
				  | SOME lst => unionPt [vars, lst])

	    (* Collect all the pending errors and return the master error string, or NONE *)
	    fun processErrors() =
		((case !multiplyAssignedNormal of
		      SOME vars => addError("Multiply assigned normal variables:\n\n  "
					    ^(strlist2str "\n  " (List.map pt2string vars)))
		    | NONE => ());
		 (case !multiplyAssignedNext of
		      SOME vars => addError("Multiply assigned next variables:\n\n  "
					    ^(strlist2str "\n  " (List.map pt2string vars)))
		    | NONE => ());
		 (case !multiplyAssignedInit of
		      SOME vars => addError("Multiply assigned init variables:\n\n  "
					    ^(strlist2str "\n  " (List.map pt2string vars)))
		    | NONE => ());
		 (case !normalAndNext of
		      SOME vars => addError
			  ("These variables appear in both normal and next assignments:\n\n  "
			   ^(strlist2str "\n  " (List.map pt2string vars)))
		    | NONE => ());
		 (case !normalAndInit of
		      SOME vars => addError
			  ("These variables appear in both normal and init assignments:\n\n  "
			   ^(strlist2str "\n  " (List.map pt2string vars)))
		    | NONE => ());
		 (case !initButNotNext of
		      SOME vars => addError
			  ("These variables appear in init, but not in next assignments:\n\n  "
			   ^(strlist2str "\n  " (List.map pt2string vars)))
		    | NONE => ());
		 (case !nextButNotInit of
		      SOME vars => addError
			  ("These variables appear in next, but not in init assignments:\n\n  "
			   ^(strlist2str "\n  " (List.map pt2string vars)))
		    | NONE => ());
		 (case !unassigned of
		      SOME vars => addError
			  ("These variables are unassigned:\n\n  "
			   ^(strlist2str "\n  " (List.map pt2string vars)))
		    | NONE => ());
		 !error)
		 
	    (* Given the list of lists of vars, find all vars that
	       appear in more than one list *)
	    fun repeats [] = []
	      | repeats [_] = []
	      | repeats (l::lst) =
		let fun ff x = List.exists(fn l => List.exists(fn y=> vtEqName(x,y)) l) lst
		    val vars = List.filter ff l
		in vars@(repeats lst)
		end
	    (* Sequent Delta |-_1 gamma *)
	    fun findStateVars (Appl(_,v, d)) =
	      let 
		val primV = findStateVars v
	      in
		if isValue (d) then 
		  List.map (fn x => Appl (dp, x, d)) primV
		else
		  let
		    val t = getExprType findObject d
		    val tvals = (getTypeValues options lim t)
		  in
		    case tvals of
		      SOME lv => List.concat
			(List.map (fn x => (List.map 
					    (fn y => 
					     Appl(dp, x, y)) lv))
			 primV)
		    | NONE => primV
		  end
	      end
	      | findStateVars (ExtractRecord (fname, v)) =
	      let
		val primV = findStateVars v
	      in
		List.map (fn x => ExtractRecord (fname, x)) primV
	      end
	      | findStateVars v = [v]
	    fun wrap (x, NormalFlag) = NormalVar x
	      | wrap (x, NextFlag) = NextVar x
	      | wrap (x, InitFlag) = InitVar x

            (* Split the type `t' of expression `e' assinged to
	       `delta' in the assignment type given by `flag', update
	       the hash of primitive vars, and return the list of all
	       primitive vars UNwrapped.  If `doUpdateCone' is true,
	       also update the cone. *)

	    fun splitType doUpdateCone (delta, t, e, flag) = 
		let val _ = debug "splitType"
		    (fn ()=>"\nsplitType("^(pt2stringDebug delta)^", "^(pt2string t)
		     ^", "^(pt2string e)^")[\n")
		    val splitType = splitType doUpdateCone
		    val res =
	      case t of
		TupleType (_, tlist) => 
		  let
		      (* List of (tau_n, #n(delta), #n(e)) for all n *)
		    val triple = List.map (fn n => (List.nth (tlist, n),
						    evaluateExpr options findObject
						    (ExtractTuple (n, delta)),
						    evaluateExpr options findObject
						    (ExtractTuple (n, e))))
		      (List.tabulate (List.length tlist, fn i => i))
		    val Ns = List.map (fn (tn, deltan, en) => 
				       splitType (deltan, tn, en, flag)) triple
		  in
		    (addPrim (delta, List.map #2 triple);
		     unionPt Ns)
		  end
	      | RecordType (_, rlist) => 
		  let
		      fun makeParts(RecordField{name=n,Type=tp}) = 
			  (tp,
			   evaluateExpr options findObject (ExtractRecord (n, delta)),
			   evaluateExpr options findObject (ExtractRecord (n, e)))
			| makeParts x = raise SympBug
			  ("collectAsstVars/splitType/parts: not RecordField: "
			   ^(pt2string x))
		      (* List of (e_n, delta.f_n, e.f_n) *)
		    val parts = List.map makeParts rlist
		    val Ns = List.map (fn (tn, deltan, en) => 
				       splitType (deltan, tn, en, flag)) parts
		  in
		    (addPrim (delta, List.map #2 parts);
		     unionPt Ns)
		  end
	      | FunType (_, t1, t2) => 
		  let
		    val dvals = getTypeValues options lim t1
		  in
		    (case dvals of
		       SOME lv => 
			 let (* List of (delta d, e d) *)
			   val pair = List.map (fn d => (evaluateExpr 
							 options findObject 
							 (Appl (dp, delta, d)),
							 evaluateExpr
							 options findObject
							 (Appl (dp, e, d))))
			     lv
			   val Ns = List.map (fn (deltan, en) => 
					      splitType (deltan, t2, en, flag)) 
			     pair
			 in		   
			   (addPrim (delta, List.map #1 pair);
			    unionPt Ns)
			 end
		     | NONE => (addPrim (delta, []);
				if doUpdateCone then 
				    (updateConeDestructive options findObject lim pvars cone 
				     (vtWrap flag delta, e); ())
				else ();
				[delta]))
		  end
	      (* Exactly the same as FunType *)
	      | ArrayType (_, t1, t2) => 
		  let
		    val dvals = getTypeValues options lim t1
		  in
		    (case
		       dvals of
		       SOME lv => 
			 let (* List of (delta d, e d) *)
			   val pair = List.map (fn d => (evaluateExpr 
							 options findObject 
							 (Appl (dp, delta, d)),
							 evaluateExpr
							 options findObject
							 (Appl (dp, e, d))))
			     lv
			   val Ns = List.map (fn (deltan, en) => 
					      splitType (deltan, t2, en, flag))
			     pair
			 in	
			   (addPrim (delta, List.map #1 pair);
			    unionPt Ns)
			 end
		     | NONE => (addPrim (delta, []);
				if doUpdateCone then
				    (updateConeDestructive options findObject lim pvars cone
				     (vtWrap flag delta, e); ())
				else ();
				[delta]))
		  end
	      | TypeInst (_, args, TypeClosure {name = name, uname = uname,
						params = params, def = def, 
						recursive = recursive,
						parent = parent}) =>
		  let
		    val nt = TypeClosure {name = name, uname = uname, 
					  params = [],
					  def = instantiateType (args, params) def,
					  recursive = recursive,
					  parent = parent}
		  in
		    splitType (delta, nt, e, flag)
		  end
	      | tp as EnumType (_, clist) => 
		  let
		    fun extractor (a as TypeConstr{Type=tp,...}) =
			  (case tp of
			       FunType _ => SOME(fn x=> ExtractAppl(a, x))
			     | _ => NONE)
		      (* This shouldn't happen anymore *)
(*  		      | extractor (f as Of(_,c,t1)) = *)
(*  			SOME(fn x=> *)
(*  			     ExtractAppl (TypeConstr{name=c, *)
(*  						     uname=c, *)
(*  						     Type=FunType(dp,t1,tp)}, *)
(*  					  x)) *)
		      (* This shouldn't happen anymore *)
		      | extractor a = raise SympBug
				 ("collectAsstVars/splitType: not a TypeConstr:\n  "
				  ^(pt2string a))
(*  			     if isName a then *)
(*  				 SOME(fn x=>TypeConstr{name=a, *)
(*  						       uname=a, *)
(*  						       Type=tp}) *)
(*  			     else raise SympBug *)
(*  				 ("collectAsstVars/splitType: not a name/type constructor:\n  " *)
(*  				  ^(pt2string a)) *)
		    (* Lits of (C_n, EC_n(delta), EC_n e) for C_n's with arguments,
		       and if `e' is not a conflicting constructor. *)
		    val triple : (ParseTree * ParseTree * ParseTree) list
		      = List.mapPartial (fn n => 
					   (Option.map(fn extr=>
						      (n, extr delta,
						       evaluateExpr options findObject (extr e)))
					             (extractor n))
			  (* if the top type constructor is known for the expression,
			     we only need that constructor.  The rest will raise
			     EvalError and will be ignored. *)
					   handle EvalError _ => NONE)
			clist
		    val innerX = 
			(case getPrimitiveIndexVar pvars delta of
			     SOME x => x
			   | NONE =>
			      let val un = Id(dp,"\""^(newName())^(pt2string delta)^"\"")
				  val x = StateVar {name = un, uname = un, 
						    Type = AbstractType clist,
						    id = un}
			      in 
				 addPrimitiveIndexVarDestructive pvars delta x;
				 x
			      end)
		    val X = vtWrap flag innerX
		    val Ns = [innerX] 
			     :: (List.map (fn (cn, deltan, en) => 
					   splitType (deltan,
						      getExprType findObject deltan,
						      en, 
						      flag)) triple)
		  in
		    (addPrim (delta, innerX :: List.map #2 triple);
		     if doUpdateCone then 
			 (addToCone (X, varsFrom options findObject pvars lim e); ())
		     else ();
		     unionPt Ns)
		  end
	      | t as (TypeClosure {def = newt, recursive = r,...}) =>
		  if r then 
		    (addPrim(delta, []);
		     if doUpdateCone then 
			 (updateConeDestructive options findObject lim pvars cone
			  (vtWrap flag delta, e); ())
		     else ();
		     [delta])
		  else
		    splitType (delta, newt, e, flag)
	      | t as Uid _ => 
		    (case findObject t of
			 SOME newtp => splitType(delta, newtp, e, flag)
		       | NONE => 
			   (addPrim(delta, []);
			    if doUpdateCone then 
				(updateConeDestructive options findObject lim pvars cone
				 (vtWrap flag delta, e); ())
			    else ();
			    [delta]))
	      | _ => (addPrim(delta, []);
		      if doUpdateCone then 
			  (updateConeDestructive options findObject lim pvars cone
			   (vtWrap flag delta, e); ())
		      else ();
		      [delta])
			 (* end of case *)
		    val _ = debug "splitType"
			(fn ()=>"splitType =\n  "^(ptlist2strDebug ",\n  " res)^"]\n")
		in res
		end  (* end of splitType *)
	    fun findPrimitiveCommon doUpdateCone (gamma, e, flag) =
	      let
		val svs = findStateVars gamma
		val svst = ((*  addPrim (gamma, svs); *)
			    List.map (fn s => (s, getExprType findObject s)) svs)
(*		val _ = print ("\nfoo: " ^ (pt2string gamma) ^ "\n")
		val _ = List.map (fn (s1, s2) => print (" " ^  (pt2string s1) 
							^ " : " ^ 
							(pt2string s2) ^
							"\n"))
		  svst *)
		fun ff(s, t) = splitType doUpdateCone (s, t, e, flag)
		val Ns = List.map ff svst
	      in
		unionPt Ns
	      end
	    fun findPrimitive (gamma, e, flag) = 
		  List.map (vtWrap flag) (findPrimitiveCommon true (gamma, e, flag))

	    fun handleConds [] = []
	      | handleConds ((CondAsst (_, e, AL)) :: rest) =
	      let
		val subtree = collectlist AL
		val CLNI = getAsstVars subtree
	      in
		(addToCones (CLNI, varsFrom options findObject pvars lim e);
		 (e, subtree) :: (handleConds rest))
	      end
	      | handleConds (x::_) = raise SympBug 
	         ("collectAsstVars/handleConds: Not CondAsst: "
		  ^(pt2stringDebug x))
	    and handleCases [] = [] 
	      | handleCases ((ChoiceAsstClosure{pattern=pat,
						uname=uname,
						names=names,
						body=AL,...}) :: rest) =
	      let
		val subtree = collectlist AL
		val CLNI = getAsstVars subtree
		val newClosure = 
		      ChoiceAsstClosure{pattern=pat,
					uname=uname,
					names=names,
					(* Erase unneeded fields *)
					body=[], parent=Fake}
	      in
		 (newClosure, subtree) :: (handleCases rest)
	      end
	      (* | handleCases ((ChoiceAsst(p,_,AL)) :: rest) =
	      let
		val subtree = collectlist AL
		val CLNI = getAsstVars subtree
	      in
		 subtree :: (handleCases rest)
	      end *)
	      | handleCases (x::_) = raise SympBug 
	        ("collectAsstVars/handleCases: Not ChoiceAsstClosure: "
		 ^(pt2stringDebug x))

	    (* Sequent N; L; I |- A (single assts) *)
	    and collect (a as (InitAsst (_, gamma, e))) =
		  let 
		    val ground = findPrimitive (gamma, e, InitFlag)
		    val asstvars = {init = ground, next = [], norm = []}
		  in
		    InitAsstTree (asstvars, gamma, e)
		  end
	      | collect (a as (NormalAsst (_, gamma, e))) =
		  let 
		    val ground = findPrimitive (gamma, e, NormalFlag)
		    val asstvars = {init = [], next = [], norm = ground}
		  in
		    NormalAsstTree (asstvars, gamma, e)
		  end
	      | collect (a as (NextAsst (_, gamma, e))) =
		  let 
		    val ground = findPrimitive (gamma, e, NextFlag)
		    val asstvars = {init = [], next = ground, norm = []}
		  in
		    NextAsstTree (asstvars, gamma, e)
		  end
	      | collect (a as (LetAsstClosure {locals = locals,
					       parent = parent,
					       body = body})) =
		  let
		    val subtree = collectlist body
		    val CLNI = getAsstVars subtree
		  in
		    LetAsstTree (CLNI, locals, subtree)
		  end
	      | collect (a as (IfAsst (_, condassts, elseasst))) =
		  let
		    val pairs = handleConds condassts
		    val last = collectlist elseasst
		    val U = unionAV (List.map (fn(_,t) => getAsstVars t) pairs)
		  in
		    IfAsstTree (U, pairs, last)
		  end
	      | collect (a as (CaseAsst (_, e, caseassts))) =
		  let
		    val pairs = handleCases (caseassts)
		    val U = unionAV (List.map (fn(_,t) => getAsstVars t) pairs)
		    val _ = addToCones (U, varsFrom options findObject pvars lim e)
		  in
		    CaseAsstTree (U, e, pairs)
		  end
	      | collect (a as (ChooseAsst (_, choiceassts))) =
		  let
		    val pairs = handleCases (choiceassts)
		    val U = unionAV (List.map (fn(_,t) => getAsstVars t) pairs)
		  in
		    ChooseAsstTree (U, pairs)
		  end
	      | collect (a as Nop _) = NopAsstTree{init = [], next = [], norm=[]}
	      | collect x = raise SympBug
		  ("collectAsstVars/collect: not an assignment:\n  "
		   ^(pt2string x))

	    (* Sequent N; L; I |- AL *)
	    and collectlist [] = 
		let val asstvars = {norm=[],next=[],init=[]}
		in NopAsstTree asstvars
		end
	      | collectlist lst = 
		let fun loop [] = []
		      | loop (a::lst) = (collect a)::(loop lst)
		    (* Collect the list of lists of variables for the list of asst trees *)
		    fun getVars ([], norms, nexts, inits) = (norms, nexts, inits)
		      | getVars (t::lst, norms, nexts, inits) =
			let val {norm=norm, next=next, init=init} = getAsstVars t
			in getVars (lst, norm::norms, next::nexts, init::inits)
			end
		    val treeList = loop lst
		    val (norms, nexts, inits) = getVars(treeList, [], [], [])
		    val norm = unionV norms
		    val next = unionV nexts
		    val init = unionV inits
		    val U = {norm = norm, next = next, init = init}
		    (* Circularity is checked at the leaves, check only multiplicity *)
		    val _ = (case repeats norms of
				 [] => ()
			       | vars => addErrorVars multiplyAssignedNormal 
				            (List.map vtName vars))
		    val _ = (case repeats nexts of
				 [] => ()
			       | vars => addErrorVars multiplyAssignedNext 
				            (List.map vtName vars))
		    val _ = (case repeats inits of
				 [] => ()
			       | vars => addErrorVars multiplyAssignedInit 
				            (List.map vtName vars))
		    val _ = (case repeats [norm, next] of
				 [] => ()
			       | vars => addErrorVars normalAndNext 
				            (List.map vtName vars))
		    val _ = (case repeats [norm, init] of
				 [] => ()
			       | vars => addErrorVars normalAndInit 
				            (List.map vtName vars))
		in 
		    case treeList of
			[] => NopAsstTree{init = [], next = [], norm=[]}
		      | [t] => t
		      | _ =>  ListAsstTree (U, treeList)
		end

            (* Remove vars in `V' from `U' and return the result *)
	    fun removeVars(U,V) =
		let fun rem(l1,l2) = List.filter(fn x=> not(List.exists(fn y=> vtEq(x,y)) l2)) l1
		    val {norm=normU, next=nextU, init=initU} = U
		    val {norm=normV, next=nextV, init=initV} = V
		in {norm=rem(normU, normV),
		    next=rem(nextU, nextV),
		    init=rem(initU, initV)}
		end
	    (* Balance the vars in the conditional branches. *)
	    fun balance tree =
		let val emptyU = {norm=[], next=[], init=[]}: AsstVars
		    fun isEmpty({norm=[], next=[], init=[]}: AsstVars) = true
		      | isEmpty _ = false
		    fun V2str{norm=norm, next=next, init=init} =
			("["^(vtlist2str "," norm)^"; "
			 ^(vtlist2str "," next)^"; "
			 ^(vtlist2str "," init)^"]")
		    fun doPair V (p,t) =
			let val (V', t') = loop(V, t)
			    val newt = 
				 if isEmpty V' then t'
				 else ListAsstTree(V, [t', NopAsstTree V'])
			in
			    (p, newt)
			end
		    (* loop returns a pair (U', tree'), where U' = U - vars(tree) *)
		    and loop (U, tree as NormalAsstTree(V,_,_)) = (removeVars(U,V), tree)
		      | loop (U, tree as NextAsstTree(V,_,_)) = (removeVars(U,V), tree)
		      | loop (U, tree as InitAsstTree(V,_,_)) = (removeVars(U,V), tree)
		      (* Adjust `nop' to assign all the variables in `U' *)
		      | loop (U, NopAsstTree _) = (emptyU, NopAsstTree U)
		      | loop (U, ListAsstTree(V, lst)) =
			  (removeVars(U,V), ListAsstTree(V, loopList(V, lst)))
		      | loop (U, t as LetAsstTree(V, defs, tree)) =
			  let val (V', tree') = loop(V, tree)
			  in if isEmpty V' then
			      (removeVars(U,V), LetAsstTree(V, defs, tree'))
			     else raise SympBug
			       ("TransPrep/balance: LetAsst didn't use these vars:\n  "
				^(V2str V')^"\nin the following assignment:\n  "
				^(avt2str t))
			  end
		      | loop (U, CaseAsstTree(V, sel, pairs)) =
			  (removeVars(U, V), CaseAsstTree(V, sel, List.map (doPair V) pairs))
		      | loop (U, IfAsstTree(V, pairs, last)) =
			  let val (_, last') = doPair V (Fake, last)
			  in 
			      (removeVars(U, V),
			       IfAsstTree(V, List.map (doPair V) pairs, last'))
			  end
		      | loop (U, ChooseAsstTree(V,pairs)) =
			  (removeVars(U, V), ChooseAsstTree(V, List.map (doPair V) pairs))
		    and loopList(V, []) = 
			  if isEmpty V then []
			  else [NopAsstTree V]
		      | loopList(V, a::lst) =
			  let val (V', a') = loop(V, a)
			  in 
			      a'::(loopList(V', lst))
			  end
		    val (_, tree') = loop(getAsstVars tree, tree)
		in
		    tree'
		end

	    (* Auxiliary function to collect the stateVar's set of
	       primitive variables.  The expression is given as `Fake'
	       to raise an exception if it's used in this case (that
	       is, it shouldn't be used). *)

	    fun processStateVar v = findPrimitiveCommon false (v, Fake, NormalFlag)
	    val primVars = List.foldr (op @) [] (List.map processStateVar stateVars)
	    val asstTree = collectlist assts
	    val {norm=normV, next=nextV, init=initV} = getAsstVars asstTree
	    val norm = List.map vtName normV
	    val next = List.map vtName nextV
	    val init = List.map vtName initV
	    val _ = (case setMinus ptEq (next, init) of
			 [] => ()
		       | vars => addErrorVars nextButNotInit vars)
	    val _ = (case setMinus ptEq (init, next) of
			 [] => ()
		       | vars => addErrorVars initButNotNext vars)
	    val assignedVars = unionPt [norm, next, init]
	    val _ = (case setMinus ptEq (primVars, assignedVars) of
			 [] => ()
		       | vars => addErrorVars unassigned vars)
	    val _ = debug "collectAsstVars"
		(fn()=>"\ncollectAsstVars => Assignment Tree:\n"
		 ^(avt2str asstTree)
		 ^"\n\nCone of Influence:\n"
		 ^(cone2str cone)
		 ^"\n\nPrimitive Variables:\n"
		 ^(primVars2strDebug pvars)^"]\n")
	in 
	    case processErrors() of
		NONE => (balance asstTree, cone, pvars)
	      | SOME str => raise TransError str
	end

    (* Later we either extend the functions above, or include more
       functions, to compute an abstracted transition relation,
       relative to a provided `Abstraction'. *)

  end
