functor TransSpecialFun(structure TransCommon: TRANS_COMMON): TRANS_SPECIAL =

  struct

    structure TransCommon = TransCommon

    open TransCommon

    open Hash
    open Pos
    open Evaluate
    open Abstract
    open ParserDefault
    open ParseTreeStruct
    open Interface
    open Options

    (* Customize an AsstVarsTree to a particular property under a
       particular abstraction *)

    (* For now, balance the variables on the conditional branches and
    add `nop's where necessary *)
    fun customizeAsstVars0 options findObject prims cone limit specList abs atree =
	let val evalExpr = evaluateExpr options findObject
	    val varsFrom = varsFrom options findObject
	    val debug = lazyVerbDebug options "customizeAsstVars"
	    fun listDiff lst1 lst2 = 
	          List.filter(fn x=>not(List.exists(fn y=>vtEq(x,y)) lst2)) lst1
	    fun doList {norm=norm, next=next, init=init} (h::t) =
	         let val {norm=hnorm, next=hnext, init=hinit} = getAsstVars h
		     val newvars = {norm=listDiff norm hnorm,
				    next=listDiff next hnext,
				    init=listDiff init hinit}
		 in doList newvars t
		 end
	      | doList vars [] = vars
            fun looppairs vars lst = List.map(fn(e,t)=>(e, loop vars t)) lst
	    and loop vars (t as (ListAsstTree(_,lst))) =
		 let val newList = List.map(fn t=>loop(getAsstVars t) t) lst
	         in (case doList vars lst of
			 {norm=[],next=[],init=[]} => ListAsstTree(vars, newList)
		       | remains =>
			     ListAsstTree(vars, (NopAsstTree remains)::newList))
		 end
	      | loop vars (LetAsstTree(_, defs, t)) = LetAsstTree(vars, defs, loop vars t)
	      | loop vars (CaseAsstTree(_, sel, lst)) =
		 CaseAsstTree(vars, sel, looppairs vars lst)
	      | loop vars (IfAsstTree(_, lst, last)) =
		 IfAsstTree(vars, looppairs vars lst, loop vars last)
	      | loop vars (ChooseAsstTree(_, lst)) = ChooseAsstTree(vars, looppairs vars lst)
	      | loop vars (NopAsstTree _) = NopAsstTree vars
	      (* Assume that other atomic assignments don't need adjustment *)
	      | loop _ t = t
	    fun isDelayed (DelayedVar x) = true
	      | isDelayed _ = false
	    val specVars = List.map cvValue (unionC (List.map (varsFrom prims limit) specList))
	    val _ = if (List.null specVars) then 
		      (lazyVerbDebug options "varsSpec" (fn () => "varsFrom failed!\n"))
		    else ()
	    fun coneFromVar v = 
		let val vType = getExprType findObject v
		    fun loop (TupleType (_, tlist)) =
			let val items = List.map (fn n => evalExpr(ExtractTuple (n, v)))
					(List.tabulate (List.length tlist, fn i => i))
			in
			    unionC (List.map coneFromVar items)
			end
		      | loop (RecordType (_, rlist)) =
			let fun makeParts(RecordField{name=n, ...}) = 
			         evalExpr (ExtractRecord (n, v))
			      | makeParts x = raise SympBug
				 ("trans_special.sml: Record not RecordField: "
				  ^(pt2string x))
			    val items = List.map makeParts rlist
			in
			    unionC (List.map coneFromVar items)
			end
		      | loop (t as FunType (_, t1, t2)) =
			let val abst = getTypeAbstraction options abs t1
			    val dlist =
				(case abst of
				     SOME t1' => getTypeValues options limit (getAbsType t1')
				   | NONE => getTypeValues options limit t1)
			in
			    (case dlist of
				 SOME ds =>
				     unionC (List.map coneFromVar
					     (List.map (fn d => evalExpr(Appl (dp, v, d))) ds))
			       | NONE => raise ProverError
				     ("Infinite type while generating transition relation:\n   "
				      ^(pt2string t)))
			end
		      (* Exactly the same as FunType *)
		      | loop (t as ArrayType (_, t1, t2)) =
			let val abst = getTypeAbstraction options abs t1
			    val dlist =
				(case abst of
				     SOME t1' => getTypeValues options limit (getAbsType t1')
				   | NONE => getTypeValues options limit t1)
			in
			    (case dlist of
				 SOME ds =>
				     unionC (List.map coneFromVar
					     (List.map (fn d => evalExpr(Appl (dp, v, d))) ds))
			       | NONE => raise ProverError
				     ("Infinite type while generating transition relation:\n   "
				      ^(pt2string t)))
			end
		      | loop (EnumType (_, clist)) =
			let fun extractor (a as TypeConstr{Type=tp,...}) =
			         (case tp of
				      FunType _ => SOME(fn x=> ExtractAppl(a, x))
				    | _ => NONE)
			      | extractor a = raise SympBug
				 ("trans_special EnumType: not a TypeConstr:\n  "
				  ^(pt2string a))
			    val extracted =
				List.mapPartial (fn n => (Option.map(fn extr =>
								     evalExpr (extr v))
							  (extractor n))
						 handle EvalError _ => NONE)
				clist
			    val innerX = 
			      (case getPrimitiveIndexVar prims v of
				   SOME x => x
				 | NONE => raise SympBug 
				      ("trans_special:  EnumType index should "
				       ^"already be in the cone.")
				     (*
				      let val un = Id(dp,newName())
					val x = StateVar {name = un, uname = un, 
					  Type = AbstractType clist,
					  id = un}
				      in 
					addPrimitiveIndexVarDestructive prims v
					x;
					x
				      end
				      *)
				     )
			in 
			    unionC (List.map coneFromVar (innerX :: extracted))
			end
		      | loop (TypeInst (_, args, TypeClosure parms)) =
			let val {name = name, 
				 uname = uname,
				 params = params, 
				 def = def, 
				 recursive = recursive,
				 parent = parent} = parms
			    val nt = TypeClosure {name = name, uname = uname, 
						  params = [],
						  def = instantiateType 
						  (args, params) def,
						  recursive = recursive,
						  parent = parent}
			in
			    loop nt
			end
		      | loop (TypeClosure {def = newt, recursive = r,...}) =
			  if r then raise SympBug 
			      "trans_special: Recursive types not yet supported"
			  else loop newt
		      | loop (vt as Uid _) =
			  (case findObject vt of
			       SOME newt => loop newt
			     | NONE => raise SympBug 
				 "trans_special:  Uid not defined")
		      | loop t =
			let val abst = getTypeAbstraction options abs t
			    fun ff f = (getVarCone options findObject prims
					limit cone (vtWrap f v))
			    val vCone = unionC (List.map ff [NormalFlag, InitFlag,NextFlag])
			    fun doAbst(SOME t') =
				let val (delayed, nonDelayed) = List.partition isDelayed vCone
				    val forceDelayed = unionC 
					  (List.map ((varsFrom prims limit) o cvValue) delayed)
				    val resultForced = unionC 
					  (List.map (coneFromVar o cvValue) forceDelayed)
				in
				    unionC [nonDelayed, resultForced]
				end
			      | doAbst NONE = vCone
			in
			    doAbst abst
			end
		  in
		    loop vType
		  end (* coneFromVar *)

	    val varsSpec = List.map cvValue (unionC (List.map coneFromVar specVars))
	    val _ = debug(fn()=>"\ncustomizeAsstVars: varsSpec = ["
			  ^(ptlist2str ", " varsSpec)^"]\n")
	    fun inC (v2, v1) = 
	      let 
		val v1 = stripObjectInst v1
		val v2 = stripObjectInst v2
		val norm = (inConeOf options findObject prims limit cone 
			    (vtWrap NormalFlag v1) (PVar v2))
		val next = (inConeOf options findObject prims limit cone
			   (vtWrap NextFlag v1) (PVar v2))
		val init = (inConeOf options findObject prims limit cone 
			    (vtWrap InitFlag v1) (PVar v2))
		val r = (case (norm, next, init) of 
			     (SOME true, _, _) => true
			   | (_, SOME true, _) => true
			   | (_, _, SOME true) => true
			   | _ => false)
	      in
		  r before (lazyVerbDebug options "inC" 
			    (fn () => if r then "TRUE\n" else "FALSE\n"))
	      end
		    
	    (* Insert variables from `vars' in place of relevant AsstVars vars *)
	    fun filterRelevant vars ({norm = norm, next = next, init = init}: AsstVars) =
		let fun doVar (vts, flag) v =
		        if List.exists (fn x => svPartOf (v, vtName x)) vts then
			      SOME(vtWrap flag v)
			else NONE
		    fun doVars(vts, flag) = List.mapPartial (doVar(vts, flag)) vars
		in
		    {norm = doVars(norm, NormalFlag),
		     next = doVars(next, NextFlag),
		     init = doVars(init, InitFlag)}
		end
	    fun isEmptyVars {norm = [], next = [], init = []} = true
	      | isEmptyVars _ = false
	    fun nonEmpty (vars: ParseTree list) lst = List.mapPartial(trim vars) lst
	    and nonEmptyNop vars t =
		(case trim vars t of
		     SOME r => r
		   | NONE => NopAsstTree{norm = [], next = [], init = []})
	    and nonEmptyPairsNop (vars: ParseTree list) lst =
	      let fun ff (e,t) = (e, nonEmptyNop vars t)
	      in
		  List.map ff lst
	      end
	    and nonEmptyPairs (vars: ParseTree list) lst =
	      let fun ff(e, t) = Option.map(fn r => (e, r))(trim vars t)
	      in
		  List.mapPartial ff lst
	      end
	    and trim vars tree = 
		let fun filterU U =
		        let val V = filterRelevant vars U
			in
			    if (isEmptyVars V) then NONE
			    else SOME V
			end
		    val doPairs = nonEmptyPairs vars
		    fun loop (NormalAsstTree(U, name, e)) =
			  Option.map(fn V=> NormalAsstTree(V, name, e)) (filterU U)
		      | loop (NextAsstTree(U, name, e)) =
			  Option.map(fn V=> (NextAsstTree(V, name, e))) (filterU U)
		      | loop (InitAsstTree(U, name, e)) =
			  Option.map(fn V=> (InitAsstTree(V, name, e))) (filterU U)
		      | loop (NopAsstTree U) =
			  Option.map(fn V=> (NopAsstTree V)) (filterU U)
		      | loop (ListAsstTree(U, atreel)) =
			  Option.map(fn V=> (ListAsstTree(V, nonEmpty vars atreel))) (filterU U)
		      | loop (CaseAsstTree(U, sel, atreel)) =
			  let fun ff V =
			      let val pairs = nonEmptyPairsNop vars atreel
			      in CaseAsstTree(V, sel, pairs)
			      end
			  in
			      Option.map ff (filterU U)
			  end
		      | loop (IfAsstTree(U, atreel, last)) =
			  let fun ff V = IfAsstTree(V, doPairs atreel, nonEmptyNop vars last)
			  in
			      Option.map ff (filterU U)
			  end
		      | loop (ChooseAsstTree(U, atreel)) =
			  Option.map(fn V=> (ChooseAsstTree(V, doPairs atreel))) (filterU U)
		      | loop (LetAsstTree(U, defs, atree)) =
			  Option.map(fn V=> 
				     (LetAsstTree(V, defs,
						  nonEmptyNop vars atree))) (filterU U)
		in
		    loop tree
		end (* trim *)

	    val trimmed = (case trim varsSpec atree of
			       SOME tree => tree
			     | NONE => NopAsstTree {norm=[], next=[], init=[]})
	    val _ = debug(fn()=>"\ncustomizeAsstVars: trimmed tree =\n"
			  ^(avt2str trimmed))
	    val result = loop (getAsstVars trimmed) trimmed
	in 
	  (varsSpec, result)
	end

    fun customizeAsstVars options findObject prims cone limit specList abs atree =
	let val (_, res) = (customizeAsstVars0 options findObject prims
			    cone limit specList abs atree)
	in
	    res
	end

    (* Perform COI reduction on the model, along with other necessary
       required transformations before generating the transition
       relation *)
    fun specializeTransRel options findObject abs trans specs =
	let val { limit=lim,...} = options
	    val stateVars = ref[]
	    fun loop (TransAtomic(AtomicModel m)) =
		let val { name=name,
			  uname=uname,
			  assts=assts,
			  cone=cone,
			  pvars=pvars,
			  absModules=absModules } = m
		    val (vars, newAssts) = 
			  customizeAsstVars0 options findObject pvars cone lim specs abs assts
		    val _ = stateVars := union ptEq [vars, !stateVars]
		in
		    TransAtomic(AtomicModel{ name=name,
					     uname=uname,
					     assts=newAssts,
					     cone=cone,
					     pvars=pvars,
					     absModules=absModules })
		end
	      | loop (TransSync2(t1, t2)) = TransSync2(loop t1, loop t2)
	      | loop (TransAsync2(t1, t2)) = TransAsync2(loop t1, loop t2)
	      | loop _ = raise SympBug
		  ("TransSpecial/specializeTransRel:\n  "
		   ^"sorry, closed form parallel composition is not implemented yet")
	    val newTrans = loop trans
	in
	    (!stateVars, newTrans)
	end

    fun specializeModel options (model: Model) specs =
	let val { trans=trans,
		  findObject=findObject,
		  stateVars=stateVars,
		  pvars=pvars,
		  abstraction=abs } = model
	    val (stateVars, newTrans) = specializeTransRel options findObject abs trans specs
	in 
	    { trans=newTrans,
	      findObject=findObject,
	      stateVars=stateVars,
	      pvars=pvars,
              abstraction=abs }: Model
	end
	

  end
