;;;; Major mode for SyMP error messages.

(require 'symp-common)

(defvar symp-error-buffer nil
  "The buffer for reporting an error.  This variable is buffer local,
and is normally stored in the SyMP server buffer.")

(defun symp-error-find-buffer ()
  "Find the error buffer for the current SyMP session, or create one
if not found."
  (let ((old-buffer (current-buffer))
	(process symp-server-process)
	(symp-buffer symp-server-process-buffer)
	(buffer nil))
    (if symp-buffer
	;; there is a symp server buffer, go there
	(progn
	  (set-buffer symp-buffer)
	  (if (and symp-error-buffer
		   (bufferp symp-error-buffer)
		   (buffer-live-p symp-error-buffer))
	      ;; the error buffer already exists, just take it
	      (setq buffer symp-error-buffer)
	    ;; no error buffer, create a new one
	    (progn
	      (setq buffer (generate-new-buffer "*SyMP Error*"))
	      ;; register it in the server buffer
	      (setq symp-error-buffer buffer)
	      ;; and set up the error buffer properly
	      (set-buffer buffer)
	      (setq symp-server-process process)
	      (setq symp-server-process-buffer symp-buffer)
	      (symp-error-mode))))
      ;; if there is no symp buffer, give up, leave `buffer' nil.
      )
    buffer))

(defun symp-error-display-message (message)
  "Pop up the error buffer and display the MESSAGE string in it."
  (let ((old-buffer (current-buffer))
	(err-buffer (symp-error-find-buffer)))
    (if err-buffer
	(progn
	  ;; put the message into the error buffer
	  (set-buffer err-buffer)
	  (goto-char (point-max))
	  (insert "\n")
	  (insert message)
	  ;; go back and display the error buffer in another window
	  (set-buffer old-buffer)
	  (display-buffer err-buffer t))
      ;; Ouch, no error buffer!  So do the best we can...
      (message message))))

(defun symp-error-mode ()
  "SyMP major mode for error reporting."
  (interactive)
  ;;; Define the major mode
  (setq major-mode 'symp-error-mode)
  (setq mode-name "SyMP Error"))

(provide 'symp-error-mode)
