(* Parser of the proof tree file format for the SyMP prover *)

structure ProofTreeStruct = ProofTreeStruct
structure ParseError = ParseError
open ProofTreeStruct
open ParseError

%%

%name ProofTree


%verbose
%pure
%header (functor ProofTreeLrValsFun
           (structure Token : TOKEN
            structure ParseError: PARSE_ERROR
            structure ProofTreeStruct: PROOF_TREE_STRUCT)
  : sig structure ParserData : PARSER_DATA
            where type result = ProofTreeStruct.ProofTreeType;
        structure Tokens : ProofTree_TOKENS
   end)

%eop EOF
%pos int
%noshift EOF

%term 
  STRING of string
| SYMBOL of string
| NUMBER of int
| DOT
| COMMA
| LP
| RP
| LB
| RB
| EQ
| EOF

%nonassoc STRING SYMBOL NUMBER DOT EQ
%nonassoc COMMA LP RP LB RB EOF

%nonterm begin of ProofTreeType
| subgoals of ProofTreeType list
| subgoal of ProofTreeType
| children of ProofTreeType list
| child of ProofTreeType
| rule of ProofTreeType
| ruleargs of ProofTreeType list
| rulearg of ProofTreeType
| subgoalid of ProofTreeType
| intlist of int list

%start begin

%%

begin       : subgoals (PTproof subgoals)
subgoals    : ([])
            | subgoal subgoals (subgoal::subgoals)
subgoal     : LP subgoalid RP (PTsubgoal(subgoalid, NONE, NONE))
            | LP subgoalid rule RP (PTsubgoal(subgoalid, SOME(rule), NONE))
            | LP subgoalid LP children RP RP (PTsubgoal(subgoalid, NONE, SOME children))
            | LP subgoalid rule  LP children RP RP (PTsubgoal(subgoalid, SOME rule, SOME children))
rule        : SYMBOL LP ruleargs RP (PTrule(SYMBOL, ruleargs))
ruleargs    : ([])
            | rulearg ([rulearg])
            | rulearg COMMA ruleargs (rulearg::ruleargs)
rulearg     : STRING (PTstring(STRING))
            | SYMBOL (PTsymbol(SYMBOL))
            | NUMBER (PTnumber(NUMBER))
            | LB ruleargs RB (PTlist ruleargs)
            | SYMBOL EQ rulearg (PTassoc(SYMBOL,rulearg))
subgoalid   : SYMBOL DOT intlist (PTsubgoalID(SYMBOL,intlist))
intlist     : NUMBER ([NUMBER])
            | NUMBER DOT intlist (NUMBER::intlist)
children    : ([])
            | child children (child::children)
            | child COMMA children (child::children)
child       : subgoalid (PTchild(subgoalid, NONE))
            | subgoalid LB STRING RB (PTchild(subgoalid, SOME STRING))
